dict_params = {}

import rbc_ukqcd_params.p_test

import rbc_ukqcd_params.p_24D
import rbc_ukqcd_params.p_32D

import rbc_ukqcd_params.p_32Dfine

import rbc_ukqcd_params.p_24DH

import rbc_ukqcd_params.p_16IH2

import rbc_ukqcd_params.p_24IH1
import rbc_ukqcd_params.p_24IH2
import rbc_ukqcd_params.p_24IH3
import rbc_ukqcd_params.p_24IH4

import rbc_ukqcd_params.p_32IH1
import rbc_ukqcd_params.p_32IH2
import rbc_ukqcd_params.p_32IH3

import rbc_ukqcd_params.p_32IcoarseH1

import rbc_ukqcd_params.p_32IfineH

import rbc_ukqcd_params.p_48I
import rbc_ukqcd_params.p_64I
import rbc_ukqcd_params.p_96I

import rbc_ukqcd_params.p_24IH01
import rbc_ukqcd_params.p_24IH02
import rbc_ukqcd_params.p_24IH03
import rbc_ukqcd_params.p_24IH04
import rbc_ukqcd_params.p_24IH05

import rbc_ukqcd_params.p_32IcoarseH01
import rbc_ukqcd_params.p_32IcoarseH02

import rbc_ukqcd_params.p_32IH01

from rbc_ukqcd_params.utils import *
